from typing import Iterable

import pytest

from dep_check.infra.std_lib_filter import StdLibSimpleFilter
from dep_check.models import SourceFile
from dep_check.use_cases.app_configuration import (
    AppConfiguration,
    AppConfigurationSingleton,
)

from .fakefile import FILE_WITH_LOCAL_IMPORT, FILE_WITH_STD_IMPORT, SIMPLE_FILE


@pytest.fixture(scope="session", autouse="true")
def setup_application_config() -> None:
    """
    Define application configuration.
    """
    app_configuration = AppConfiguration(std_lib_filter=StdLibSimpleFilter())
    AppConfigurationSingleton.define_app_configuration(app_configuration)


@pytest.fixture(scope="session")
def source_files() -> Iterable[SourceFile]:
    """
    Iter over test source files.
    """
    return SIMPLE_FILE, FILE_WITH_LOCAL_IMPORT, FILE_WITH_STD_IMPORT



"""
Implementations of IDependenciesPrinter
"""
from dataclasses import asdict, dataclass
from enum import Enum
from pathlib import Path
from subprocess import check_output
from sys import stdin, stdout
from typing import Dict, Iterable, Iterator, List, Optional, Tuple

import yaml
from jinja2 import Template

from dep_check.models import GlobalDependencies, Module, Rules, iter_all_modules
from dep_check.use_cases.build import IConfigurationWriter
from dep_check.use_cases.check import DependencyError, IReportPrinter
from dep_check.use_cases.draw_graph import IGraphDrawer
from dep_check.use_cases.interfaces import Configuration


class Format(Enum):
    SUCCESS = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    INFO = "\033[94m"


class YamlConfigurationIO(IConfigurationWriter):
    """
    Configuration yaml serialization.
    """

    def __init__(self, config_path: str):
        self.config_path = config_path

    def write(self, configuration: Configuration) -> None:
        if self.config_path == "-":
            stream = stdout
            stream.write("---\n\n")
            yaml.safe_dump(asdict(configuration), stream)
        else:
            with open(self.config_path, "w") as stream:
                stream.write("---\n\n")
                yaml.safe_dump(asdict(configuration), stream)

    def read(self) -> Configuration:
        if self.config_path == "-":
            return Configuration(**yaml.safe_load(stdin))

        with open(self.config_path) as stream:
            return Configuration(**yaml.safe_load(stream))


class ReportPrinter(IReportPrinter):
    """
    Print the report after checking the files
    """

    def _error(self, dep_errors: List[DependencyError]) -> None:
        """
        Log errors.
        """
        module_errors = {
            error.module: [e for e in dep_errors if e.module == error.module]
            for error in dep_errors
        }

        for module, errors in sorted(module_errors.items()):
            print("\nModule " + Format.BOLD.value + module + Format.ENDC.value + ":")
            print(
                " \u2022 "
                + Format.FAIL.value
                + "Unauthorized modules:"
                + Format.ENDC.value
            )

            for error in errors:
                print(f"\t- {error.dependency}")
            print("\n \u2022 " + Format.INFO.value + "Rules:" + Format.ENDC.value)

            for rule in errors[0].rules:
                print(f"\t- {rule}")

    def _warning(self, unused_rules: Rules) -> None:
        """
        Log warnings
        """
        previous_wildcard = ""
        for wildcard, rule in sorted(unused_rules):
            if wildcard != previous_wildcard:
                print(
                    "\nWildcard "
                    + Format.BOLD.value
                    + wildcard
                    + Format.ENDC.value
                    + ":"
                )
                print(
                    " \u2022 "
                    + Format.WARNING.value
                    + "Unused rules:"
                    + Format.ENDC.value
                )
                previous_wildcard = wildcard
            print(f"\t- {wildcard}: {rule}")

    def print_report(
        self, errors: List[DependencyError], unused_rules: Rules, nb_files: int
    ) -> None:
        """
        Print report
        """
        if errors:
            print(
                "\n\n"
                + Format.BOLD.value
                + Format.FAIL.value
                + "IMPORT ERRORS".center(30)
                + Format.ENDC.value
            )
            self._error(errors)

        if unused_rules:
            print(
                "\n\n"
                + Format.BOLD.value
                + Format.WARNING.value
                + "UNUSED RULES".center(30)
                + Format.ENDC.value
            )
            self._warning(unused_rules)

        if not errors and not unused_rules:
            print(
                Format.SUCCESS.value + "\nEverything is in order! " + Format.ENDC.value
            )
        print(
            "\n * "
            + Format.FAIL.value
            + f"{len(errors)} errors"
            + Format.ENDC.value
            + " and "
            + Format.WARNING.value
            + f"{len(unused_rules)} warnings"
            + Format.ENDC.value
            + f" in {nb_files} files."
        )


def read_graph_config(conf_path: str) -> Dict:
    """
    Used to read the graph configuration file, and make it a Dictionary
    """
    with open(conf_path) as stream:
        return yaml.safe_load(stream)


@dataclass(init=False)
class Graph:
    """
    Dataclass representing the information to draw a graph
    """

    def __init__(self, svg_file_name: str, graph_config: Optional[Dict] = None):
        self.svg_file_name = svg_file_name
        self.graph_config = graph_config or {}
        self.dot_file_name: str = "/tmp/graph.dot"
        self.node_color: str = self.graph_config.get("node_color", "white")
        self.background_color: str = self.graph_config.get("bgcolor", "transparent")
        self.layers: dict = self.graph_config.get("layers", {})


class GraphDrawer(IGraphDrawer):
    """
    Write dot / svg files corresponding to the project dependencies
    """

    def __init__(self, graph: Graph):
        self.graph = graph
        self.header = Template(
            "digraph G {\n"
            "splines=true;\n"
            "node[shape=box fontname=Arial style=filled fillcolor={{nodecolor}}];\n"
            "bgcolor={{bgcolor}}\n\n\n"
        ).render(nodecolor=self.graph.node_color, bgcolor=self.graph.background_color)
        self.body = ""
        self.footer = "}\n"

        self.subgraph = Template(
            "subgraph cluster_{{subgraph_name}} {\n"
            "node [style=filled fillcolor={{color}}];\n"
            "{{list_modules}};\n"
            'label="{{subgraph_name}}";\n'
            "color={{color}};\n"
            "penwidth=2;\n"
            "}\n\n\n"
        )

    def _iter_layer_modules(
        self, global_dep: GlobalDependencies
    ) -> Iterator[Tuple[str, Iterable[Module]]]:

        for layer in self.graph.layers:
            yield layer, [
                m
                for m in iter_all_modules(global_dep)
                if m.startswith(tuple(self.graph.layers[layer]["modules"]))
            ]

    def _write_dot(self, global_dep: GlobalDependencies) -> bool:
        if not global_dep:
            return False

        for layer, modules in self._iter_layer_modules(global_dep):

            self.body += self.subgraph.render(
                subgraph_name=layer,
                color=self.graph.layers[layer].get("color", self.graph.node_color),
                list_modules=str(modules)[1:-1].replace("'", '"'),
            )

        for module, deps in global_dep.items():
            for dep in deps:
                self.body += '"{}" -> "{}"\n'.format(module, dep.main_import)

        with open(self.graph.dot_file_name, "w") as out:
            out.write(self.header)
            out.write(self.body)
            out.write(self.footer)

        return True

    def _write_svg(self) -> None:
        svg_string = check_output(["dot", "-Tsvg", self.graph.dot_file_name]).decode()
        if self.graph.svg_file_name == "-":
            stdout.write(svg_string)
        else:
            with open(self.graph.svg_file_name, "w") as stream:
                stream.write(svg_string)

    def write(self, global_dep: GlobalDependencies):
        if Path(self.graph.svg_file_name).suffix == ".dot":
            self.graph.dot_file_name = self.graph.svg_file_name
            self._write_dot(global_dep)
        else:
            if self._write_dot(global_dep):
                self._write_svg()



"""
Define all the business models of the application.
"""

from dataclasses import dataclass, field
from typing import Dict, FrozenSet, Iterator, List, NewType, Set, Tuple

Module = NewType("Module", str)


@dataclass(frozen=True)
class Dependency:
    """
    A complete information about a dependency

    With 'from a import b, c', main_import = 'a' and sub_imports = {b, c}
    With 'import e', main_import = 'e' and sub_imports = {}
    """

    main_import: Module = Module("")
    sub_imports: FrozenSet[Module] = field(default_factory=frozenset)


Dependencies = Set[Dependency]

SourceCode = NewType("SourceCode", str)

ModuleWildcard = NewType("ModuleWildcard", str)

Rule = Tuple[ModuleWildcard, ModuleWildcard]
Rules = Set[Rule]

DependencyRules = Dict[str, List[ModuleWildcard]]

GlobalDependencies = Dict[Module, Dependencies]


def get_parent(module: Module) -> Module:
    """
    Get the parent module of a given one.
    """
    return Module(module.rpartition(".")[0])


def iter_all_modules(global_dep: GlobalDependencies) -> Iterator[Module]:
    def iter_(global_dep: GlobalDependencies) -> Iterator[Module]:
        for module, dependencies in global_dep.items():
            yield module
            yield from (d.main_import for d in dependencies)

    return iter(set(iter_(global_dep)))


@dataclass
class SourceFile:
    """
    A complete information about a source file.
    """

    module: Module
    code: SourceCode



#! /usr/bin/env python3
"""
Check dependencies of the project
"""

import argparse
import logging
import sys
from dataclasses import dataclass
from typing import Callable

from dep_check.infra.file_system import source_file_iterator
from dep_check.infra.go_parser import GoParser
from dep_check.infra.io import (
    Graph,
    GraphDrawer,
    ReportPrinter,
    YamlConfigurationIO,
    read_graph_config,
)
from dep_check.infra.python_parser import PythonParser
from dep_check.infra.std_lib_filter import StdLibSimpleFilter
from dep_check.use_cases.app_configuration import (
    AppConfiguration,
    AppConfigurationSingleton,
)
from dep_check.use_cases.build import BuildConfigurationUC
from dep_check.use_cases.check import CheckDependenciesUC
from dep_check.use_cases.draw_graph import DrawGraphUC
from dep_check.use_cases.interfaces import ExitCode

FEATURE_PARSER = argparse.ArgumentParser(description="Chose your feature")
FEATURE_PARSER.add_argument(
    "feature",
    type=str,
    help="The feature you want.",
    choices=["build", "check", "graph"],
)

BUILD_PARSER = argparse.ArgumentParser(description="Build your dependency rules")
BUILD_PARSER.add_argument(
    "build", type=str, help="The build feature.", choices=["build"]
)
BUILD_PARSER.add_argument(
    "root_dir", type=str, help="The source root dir for search and check."
)
BUILD_PARSER.add_argument(
    "-o",
    "--output",
    type=str,
    help="The name of the yaml file you want",
    default="dependency_config.yaml",
)
BUILD_PARSER.add_argument(
    "--lang",
    type=str,
    help="The language of your project",
    default="python",
    choices=["py", "python", "go", "golang"],
)


CHECK_PARSER = argparse.ArgumentParser(description="Check the dependencies")
CHECK_PARSER.add_argument(
    "check", type=str, help="The check feature.", choices=["check"]
)
CHECK_PARSER.add_argument(
    "root_dir", type=str, help="The source root dir for search and check."
)
CHECK_PARSER.add_argument(
    "-c",
    "--config",
    type=str,
    help="The name of the yaml file you want",
    default="dependency_config.yaml",
)


GRAPH_PARSER = argparse.ArgumentParser(description="Draw a dependency graph")
GRAPH_PARSER.add_argument(
    "graph", type=str, help="The graph feature.", choices=["graph"]
)
GRAPH_PARSER.add_argument(
    "root_dir", type=str, help="The source root dir for search and check."
)
GRAPH_PARSER.add_argument(
    "-o",
    "--output",
    type=str,
    help="The name of the svg/dot file you want",
    default="dependency_graph.svg",
)
GRAPH_PARSER.add_argument(
    "-c", "--config", type=str, help="The yaml file representing the graph options."
)
GRAPH_PARSER.add_argument(
    "--lang",
    type=str,
    help="The language of your project",
    choices=["py", "python", "go", "golang"],
)


class MissingOptionError(Exception):
    """
    DependencyError raised when a missing option is found.
    """


@dataclass
class Feature:
    parser: argparse.ArgumentParser
    use_case_factory: Callable


class MainApp:
    """
    Main application class.

    This class is and must be logic-less. Its only aim is to connect
    different components to each other in order to make use_cases working.
    For each abstract interface, this class selects a concrete implementation for the
    application lifetime.
    """

    def __init__(self) -> None:
        self.feature = FEATURE_PARSER.parse_args(sys.argv[1:2]).feature

        try:
            self.args = DEP_CHECK_FEATURES[self.feature].parser.parse_args()
        except KeyError:
            raise MissingOptionError()

    def main(self) -> int:
        code = ExitCode.OK
        self.create_app_configuration()
        try:
            use_case = DEP_CHECK_FEATURES[self.feature].use_case_factory(self)
            use_case.run()

        except KeyError:
            raise MissingOptionError()

        return code.value

    @staticmethod
    def create_app_configuration() -> None:
        """
        Create and set the global application configuration.
        """
        app_configuration = AppConfiguration(std_lib_filter=StdLibSimpleFilter())
        AppConfigurationSingleton.define_app_configuration(app_configuration)

    def create_build_use_case(self) -> BuildConfigurationUC:
        """
        Plumbing to make build use case working.
        """
        configuration_io = YamlConfigurationIO(self.args.output)
        code_parser = (
            PythonParser() if self.args.lang in ["py", "python"] else GoParser()
        )
        source_files = source_file_iterator(self.args.root_dir, self.args.lang[:2])
        return BuildConfigurationUC(
            configuration_io, code_parser, source_files, self.args.lang
        )

    def create_check_use_case(self) -> CheckDependenciesUC:
        """
        Plumbing to make check use case working.
        """
        configuration = YamlConfigurationIO(self.args.config).read()
        code_parser = (
            PythonParser() if configuration.lang in ["py", "python"] else GoParser()
        )
        report_printer = ReportPrinter()
        source_files = source_file_iterator(self.args.root_dir, configuration.lang[:2])
        return CheckDependenciesUC(
            configuration, report_printer, code_parser, source_files
        )

    def create_graph_use_case(self) -> DrawGraphUC:
        """
        Plumbing to make draw_graph use case working.
        """
        graph_conf = read_graph_config(self.args.config) if self.args.config else None

        if self.args.lang:
            lang = self.args.lang
        elif graph_conf and "lang" in graph_conf:
            lang = graph_conf["lang"]
        else:
            lang = "python"

        code_parser = PythonParser() if lang in ["py", "python"] else GoParser()
        source_files = source_file_iterator(self.args.root_dir, lang[:2])
        graph = Graph(self.args.output, graph_conf)
        graph_drawer = GraphDrawer(graph)
        return DrawGraphUC(graph_drawer, code_parser, source_files, graph_conf)


DEP_CHECK_FEATURES = {
    "build": Feature(BUILD_PARSER, MainApp.create_build_use_case),
    "check": Feature(CHECK_PARSER, MainApp.create_check_use_case),
    "graph": Feature(GRAPH_PARSER, MainApp.create_graph_use_case),
}


def main() -> None:
    try:
        sys.exit(MainApp().main())
    except MissingOptionError:
        logging.error(
            "You have to write which feature you want to use among [build,check,graph]"
        )
        sys.exit(1)


if __name__ == "__main__":
    main()



"""
Common use cases interfaces.
"""
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum

from dep_check.models import Dependencies, DependencyRules


@dataclass
class Configuration:
    """
    The configuration for the tools.
    """

    dependency_rules: DependencyRules = field(default_factory=dict)
    lang: str = "python"
    local_init: bool = False


class IStdLibFilter(ABC):
    """
    Filter for remove stdlib from dependencies.
    """

    @abstractmethod
    def filter(self, dependencies: Dependencies) -> Dependencies:
        """
        Remove dependencies that are part of stdlib.
        """


class ExitCode(Enum):
    """
    Describe possible exit codes
    """

    OK = 0
    KO = 1



from unittest.mock import Mock

from pytest import raises

from dep_check.use_cases.app_configuration import (
    AppConfigurationAlreadySetException,
    AppConfigurationSingleton,
)


def test_app_configuration_already_set() -> None:
    """
    Test that App configuration singleton can not be defined if already set
    """
    # Given
    AppConfigurationSingleton._already_set = True  # pylint: disable=protected-access
    config = Mock()

    # When -Then
    with raises(AppConfigurationAlreadySetException):
        AppConfigurationSingleton.define_app_configuration(config)



import ast
from typing import Any, FrozenSet, List

from dep_check.dependency_finder import IParser
from dep_check.models import (
    Dependencies,
    Dependency,
    Module,
    ModuleWildcard,
    SourceFile,
)


class _ImportVisitor(ast.NodeVisitor):
    """
    Implementation of a NodeVisitor, to scan import of a python code.
    """

    def __init__(self, current_module: str) -> None:
        self._dependencies: Dependencies = set()
        self.current_module_parts: List[str] = current_module.split(".")

    @property
    def dependencies(self) -> Dependencies:
        """
        Dependencies found during the scan.
        """
        return self._dependencies

    def visit(self, node: Any) -> None:
        modules: FrozenSet[Dependency] = frozenset()
        if isinstance(node, ast.Import):
            modules = frozenset(Dependency(Module(alias.name)) for alias in node.names)

        elif isinstance(node, ast.ImportFrom):
            module = Module(node.module or "")
            if node.level:
                parent_module = ".".join(self.current_module_parts[: -node.level])
                if node.module:
                    module = Module("{}.{}".format(parent_module, node.module))
                else:
                    module = Module(parent_module)
            modules = frozenset((Dependency(module),))
        self._dependencies |= modules

        super().visit(node)


class _ImportFromVisitor(ast.NodeVisitor):
    def __init__(self, current_module: str) -> None:
        self._dependencies: Dependencies = set()
        self.current_module_parts: List[str] = current_module.split(".")

    @property
    def dependencies(self) -> Dependencies:
        """
        Dependencies found during the scan.
        """
        return self._dependencies

    def visit(self, node: Any) -> None:
        modules: FrozenSet[Dependency] = frozenset()
        if isinstance(node, ast.Import):
            modules = frozenset(Dependency(Module(alias.name)) for alias in node.names)
        elif isinstance(node, ast.ImportFrom):
            module = Module(node.module or "")
            if node.level:
                parent_module = ".".join(self.current_module_parts[: -node.level])
                if node.module:
                    module = Module("{}.{}".format(parent_module, node.module))
                else:
                    module = Module(parent_module)
            sub_imports = frozenset(Module(alias.name) for alias in node.names)
            modules = frozenset((Dependency(module, sub_imports),))

        self._dependencies |= modules

        super().visit(node)


class PythonParser(IParser):
    """
    Implementation of the interface, to parse python
    """

    def wildcard_to_regex(self, module: ModuleWildcard) -> str:
        """
        Return a regex expression for the Module from wildcard
        """
        module_regex = module.replace(".", "\\.").replace("*", ".*")
        module_regex = module_regex.replace("[!", "[^").replace("?", ".?")

        # Special char including a module along with all its sub-modules:
        module_regex = module_regex.replace("%", r"(\..*)?$")
        return module_regex

    def find_dependencies(self, source_file: SourceFile) -> Dependencies:
        """
        Scan a python source file and return its dependencies.
        """
        visitor = _ImportVisitor(source_file.module)
        node = ast.parse(source_file.code)
        visitor.visit(node)
        return visitor.dependencies

    def find_import_from_dependencies(self, source_file: SourceFile) -> Dependencies:
        """
        Scan a python source file and return its dependencies.
        """
        visitor = _ImportFromVisitor(source_file.module)
        node = ast.parse(source_file.code)
        visitor.visit(node)
        return visitor.dependencies



"""
Tests about get_dependencies function.
"""
import re

from dep_check.dependency_finder import get_dependencies, get_import_from_dependencies
from dep_check.infra.python_parser import PythonParser
from dep_check.models import Dependency, Module, ModuleWildcard, SourceCode, SourceFile

_SIMPLE_CASE = """
import simple
import module.inside.module
from amodule import aclass
"""
_SIMPLE_RESULT = frozenset(
    (
        Dependency(Module("simple")),
        Dependency(Module("amodule")),
        Dependency(Module("module.inside.module")),
    )
)
_SIMPLE_RESULT_IMPORT_FROM = frozenset(
    (
        Dependency(Module("simple")),
        Dependency(Module("amodule"), frozenset((Module("aclass"),))),
        Dependency(Module("module.inside.module")),
    )
)

_LOCAL_CASE = """
import simple
from . import aclass
from .inside.module import aclass
"""
_LOCAL_RESULT = set(
    (
        Dependency(Module("simple")),
        Dependency(Module("module")),
        Dependency(Module("module.inside.module")),
    )
)
_LOCAL_RESULT_IMPORT_FROM = set(
    (
        Dependency(Module("simple")),
        Dependency(Module("module"), frozenset((Module("aclass"),))),
        Dependency(Module("module.inside.module"), frozenset((Module("aclass"),))),
    )
)

PARSER = PythonParser()


class TestGetDependencies:
    @staticmethod
    def test_empty() -> None:
        """
        Test empty code case.
        """
        # Given
        module = Module("")
        source_code = SourceCode("")
        source_file = SourceFile(module=module, code=source_code)

        # When
        dependencies = get_dependencies(source_file, PARSER)

        # Then
        assert dependencies == frozenset()

    @staticmethod
    def test_simple_case() -> None:
        """
        Test simple code case.
        """
        # Given
        module = Module("toto_program")
        source_file = SourceFile(module=module, code=SourceCode(_SIMPLE_CASE))

        # When
        dependencies = get_dependencies(source_file, PARSER)

        # Then
        assert dependencies == _SIMPLE_RESULT

    @staticmethod
    def test_local_import_case() -> None:
        """
        Test code with local import case.
        """
        # Given
        module = Module("module.toto")
        source_file = SourceFile(module=module, code=SourceCode(_LOCAL_CASE))

        # When
        dependencies = get_dependencies(source_file, PARSER)

        # Then
        assert dependencies == _LOCAL_RESULT


class TestGetImportFromDependencies:
    @staticmethod
    def test_empty() -> None:
        """
        Test empty code case.
        """
        # Given
        module = Module("")
        source_code = SourceCode("")
        source_file = SourceFile(module=module, code=source_code)

        # When
        dependencies = get_import_from_dependencies(source_file, PARSER)

        # Then
        assert dependencies == frozenset()

    @staticmethod
    def test_simple_case() -> None:
        """
        Test simple code case.
        """
        # Given
        module = Module("toto_program")
        source_file = SourceFile(module=module, code=SourceCode(_SIMPLE_CASE))

        # When
        dependencies = get_import_from_dependencies(source_file, PARSER)

        # Then
        assert dependencies == _SIMPLE_RESULT_IMPORT_FROM

    @staticmethod
    def test_local_import_case() -> None:
        """
        Test code with local import case.
        """
        # Given
        module = Module("module.toto")
        source_file = SourceFile(module=module, code=SourceCode(_LOCAL_CASE))

        # When
        dependencies = get_import_from_dependencies(source_file, PARSER)

        # Then
        assert dependencies == _LOCAL_RESULT_IMPORT_FROM

    @staticmethod
    def test_multi_imports_after_from() -> None:
        # Given
        module = Module("module.toto")
        source_file = SourceFile(
            module=module,
            code=SourceCode("from module import amodule, othermodule, moduleagain"),
        )

        # When
        dependencies = get_import_from_dependencies(source_file, PARSER)

        # Then
        assert dependencies == set(
            (
                Dependency(
                    Module("module"),
                    frozenset(
                        (
                            Module("amodule"),
                            Module("othermodule"),
                            Module("moduleagain"),
                        )
                    ),
                ),
            )
        )


class TestRegexToWildcard:
    """
    Test build module regex function.
    """

    @staticmethod
    def test_empty() -> None:
        """
        Test empty case.
        """
        # Given
        module = ModuleWildcard("")

        # When
        regex = PARSER.wildcard_to_regex(module)

        # Then
        assert regex == ""

    @staticmethod
    def test_simple_module() -> None:
        """
        Test simple case.
        """
        # Given
        module = ModuleWildcard("toto")

        # When
        regex = PARSER.wildcard_to_regex(module)

        # Then
        assert re.match(regex, "toto")
        assert not re.match(regex, "tata")
        assert not re.match(regex, "titi.toto")

    @staticmethod
    def test_nested_module() -> None:
        """
        Test nested case.
        """
        # Given
        module = ModuleWildcard("toto.tata")

        # When
        regex = PARSER.wildcard_to_regex(module)

        # Then
        assert re.match(regex, "toto.tata")
        assert not re.match(regex, "toto")
        assert not re.match(regex, "tata")
        assert not re.match(regex, "titi.toto")

    @staticmethod
    def test_quesiton_mark() -> None:
        """
        Test question mark case
        """
        # Given
        module = ModuleWildcard("t?to.?at?")

        # When
        regex = PARSER.wildcard_to_regex(module)

        # Then
        assert re.match(regex, "toto.tata")
        assert re.match(regex, "t2to.bato")
        assert re.match(regex, "t#to.!at&")
        assert not re.match(regex, "toto")
        assert not re.match(regex, "tata")
        assert not re.match(regex, "toti.toto")

    @staticmethod
    def test_asterisk() -> None:
        """
        Test asterisk case
        """
        # Given
        module = ModuleWildcard("toto*.*")

        # When
        regex = PARSER.wildcard_to_regex(module)

        # Then
        assert re.match(regex, "toto.tata")
        assert re.match(regex, "toto_2351.titi")
        assert re.match(regex, "toto_azerty.titi.toto.tata")
        assert not re.match(regex, "toto")
        assert not re.match(regex, "tototata")
        assert not re.match(regex, "toti.toto")

    @staticmethod
    def test_percentage() -> None:
        """
        Test percentage case
        """
        # Given
        module = ModuleWildcard("toto.tata%")

        # When
        regex = PARSER.wildcard_to_regex(module)

        # Then
        assert re.match(regex, "toto.tata")
        assert re.match(regex, "toto.tata.titi")
        assert re.match(regex, "toto.tata.titi.tutu.tototata.tititutu")
        assert not re.match(regex, "toto")
        assert not re.match(regex, "toto.tata_123")



"""
Build rules more restrictive rules from existing source use case.
"""
from abc import ABC, abstractmethod
from typing import Dict, Iterator

from dep_check.dependency_finder import IParser, get_dependencies
from dep_check.models import Dependencies, Module, ModuleWildcard, SourceFile

from .app_configuration import AppConfigurationSingleton
from .interfaces import Configuration


class IConfigurationWriter(ABC):
    """
    Interface for writing a script configuration.
    """

    @abstractmethod
    def write(self, configuration: Configuration) -> None:
        """
        Write a script configuration.
        """


class BuildConfigurationUC:
    """
    Build more restrictive rules from existing list of files.
    """

    def __init__(
        self,
        printer: IConfigurationWriter,
        parser: IParser,
        source_files: Iterator[SourceFile],
        lang: str,
    ) -> None:
        app_configuration = AppConfigurationSingleton.get_instance()
        self.std_lib_filter = app_configuration.std_lib_filter
        self.printer = printer
        self.parser = parser
        self.source_files = source_files
        self.lang = lang

    def run(self) -> None:
        """
        Build configuration from existing source files.
        """
        global_dependencies: Dict[Module, Dependencies] = {}
        for source_file in self.source_files:
            dependencies = get_dependencies(source_file, self.parser)
            dependencies = self.std_lib_filter.filter(dependencies)

            global_dependencies[source_file.module] = dependencies

        dependency_rules = {}
        for module, dependencies in global_dependencies.items():
            dependency_rules[str(module)] = [
                ModuleWildcard(dependency.main_import) for dependency in dependencies
            ]

        self.printer.write(Configuration(dependency_rules, self.lang))



import logging
from ctypes import Structure, c_char_p, c_longlong, cdll
from os import environ
from subprocess import CalledProcessError, run

from dep_check.dependency_finder import IParser
from dep_check.models import Dependencies, Dependency, ModuleWildcard, SourceFile


class GoString(Structure):  # pylint: disable=too-few-public-methods
    _fields_ = [("data", c_char_p), ("n", c_longlong)]


class GoParser(IParser):
    """
    Implementation of the interface, to parse go
    """

    def __init__(self):
        self.lib_path = environ.get("GOLIB", f'{environ["HOME"]}/.dep-check')

        try:
            self.lib = cdll.LoadLibrary(f"{self.lib_path}/go_parse.so")
        except OSError:
            logging.info("go_parse.so not found. Building...")
            try:
                run(
                    ["go", "get", "-d", "github.com/lumapps/dep-check/dep_check/lib"],
                    check=True,
                )
                run(
                    [
                        "go",
                        "build",
                        "-o",
                        f"{self.lib_path}/go_parse.so",
                        "-buildmode=c-shared",
                        "$GOPATH/src/github.com/lumapps/dep-check/dep_check/lib/go_parse.go",
                    ],
                    check=True,
                )
                environ["GOLIB"] = self.lib_path
                logging.info("GO lib go_parse.so built")

            except CalledProcessError:
                logging.error(
                    "Unable to build the GO lib from "
                    "github.com/lumapps/dep-check/dep_check/lib"
                )
                raise

            try:
                self.lib = cdll.LoadLibrary(f"{self.lib_path}/go_parse.so")
                # After building the lib, we try to load it again
            except OSError:
                logging.error("Error while loading the library. Please try again")
                raise

        self.lib.FindDependencies.argtypes = [GoString]
        self.lib.FindDependencies.restype = GoString

    def wildcard_to_regex(self, module: ModuleWildcard) -> str:
        """
        Return a regex expression for the Module from wildcard
        """
        module_regex = module.replace(".", "\\.").replace("*", ".*")
        module_regex = module_regex.replace("[!", "[^").replace("?", ".?")

        # Special char including a module along with all its sub-modules:
        module_regex = module_regex.replace("%", r"(/.*)?$")
        return module_regex

    def find_dependencies(self, source_file: SourceFile) -> Dependencies:
        if source_file.code == "":
            return set()
        deps_string = self.lib.FindDependencies(
            GoString(source_file.code.encode(), len(source_file.code))
        )
        deps_list = deps_string.data.decode("utf-8").replace('"', "").split(";")[:-1]

        dependecies: Dependencies = {Dependency(dep) for dep in deps_list}
        return dependecies

    def find_import_from_dependencies(self, source_file: SourceFile) -> Dependencies:
        return self.find_dependencies(source_file)
